/** @file
 *
 * This library contains functionality for the broker in a distributed RPC
 *   system.
 *
 * @author  Victor Rodrigues (CERN, BE-ICS-SDS)
  */

#uses "fwRPC/fwRPC_common.ctl"
#uses "fwRPC/fwRPC_msgBroker.ctl"
#uses "fwRPC/fwRPC_exception.ctl"
#uses "fwRPC/fwRPC_requestModel.ctl"
#uses "fwRPC/fwRPC_debug.ctl"
#uses "CtrlSerialization"


/**
 * Collection of cores and their available procedures.
 * Structure:
 *  DYN_DYN_MIXED:
 *    [x]: DYN_MIXED: (core)
 *           [1]: DYN_STRING: (project info)
 *                  [1]: DIST_NODE_PROJECT
 *                  [2]: DIST_NODE_NAME
 *           [2]: DYN_STRING: (function list)
 *                  [x]: procedure name
 */
private global dyn_dyn_mixed g_fwRPC_broker_cores;


/**
 * Current index in round-robin load balancing.
 */
private global int g_fwRPC_broker_coreIndex;


/**
 * Time to wait in seconds between ticks to check for core delegation timeout
 */
private global const int _fwRPC_BROKER_TICK_RATE_S = 2;


/**
 * Time to wait for a core to initially process a new request before reassigning
 * to a different core
 */
private global const int _fwRPC_BROKER_CORE_EXPIRE_TIME_S = 3;


/**
 * Mapping of pending calls. New client requests are appended to this
 * collection, signifying that they have not yet received indication that the
 * core has processed it. The broker will wait for
 * _fwRPC_BROKER_CORE_EXPIRE_TIME_S seconds before attempting to reassign the
 * call to a new core.
 *
 * Once the broker receives a status update for the given call, the call will be
 * removed from this mapping.
 *
 * key: call hdl
 * value: call packet (fwRPC_DISTMSG_REQUEST_*) with expire timestamp at index 4
 */
private global mapping _g_fwRPC_broker_pendingCalls;


//------------------------------------------------------------------------------


/** 
 * Initialize the broker module.
 *
 * @returns nothing
 *
 * @throws nothing
 */
void fwRPC_broker_initialize() {
  dynClear(g_fwRPC_broker_cores);
  mappingClear(_g_fwRPC_broker_pendingCalls);
  g_fwRPC_broker_coreIndex = 0;

  fwRPC_msgBroker_initialize();

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_DISCONNECT,
   "_fwRPC_broker_hdlDisconnect",
    fwRPC_DIST_TAG_CORE,
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_REQUEST,
    "_fwRPC_broker_hdlRequest",
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_STATUS,
    "_fwRPC_broker_hdlStatus",
    fwRPC_DIST_TAG_CORE
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_SET_RESULT,
    "_fwRPC_broker_hdlSetResult",
    fwRPC_DIST_TAG_CORE
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_GET_RESULT,
    "_fwRPC_broker_hdlGetResult",
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_REGISTER,
    "_fwRPC_broker_hdlRegister",
    fwRPC_DIST_TAG_CORE
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_ACKNOWLEDGE,
    "_fwRPC_broker_hdlAcknowledge",
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_CANCEL,
    "_fwRPC_broker_hdlCancel",
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_RECOVER,
    "_fwRPC_broker_hdlRecover",
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_LISTSVC,
    "_fwRPC_broker_hdlListServices",
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_dbg_hook(
    fwRPC_common_cleanProjectName(getSystemName()),
    fwRPC_DBG_MODULE_BROKER,
    "_fwRPC_broker_dbg_report"
  );


  int beginTick_tid = startThread("_fwRPC_broker_beginTick");
  if (beginTick_tid == -1) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Failed to start broker tick. Core delegation failure will not be detected"
    );
  }


  fwRPC_msgBroker_listen(false, "fwRPC_broker_newMbxCB");
}


/**
 * @internal
 *
 * Starts the broker loop which checks for request-core delegation timeouts.
 * Attempts to reassign the request to another core on timeout.
 *
 * @returns nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Thread}
 *
 * @throws nothing
 */
private void _fwRPC_broker_beginTick() {
  time  currentTime;
  mixed expTime;

  while(true) {
    currentTime = getCurrentTime();

    int len = mappinglen(_g_fwRPC_broker_pendingCalls);
    dyn_anytype callHdls = mappingKeys(_g_fwRPC_broker_pendingCalls);
    for (int i=1; i <= dynlen(callHdls); ++i) {
      anytype hdl = callHdls[i];
      dyn_mixed call = _g_fwRPC_broker_pendingCalls[hdl];

      expTime = call[fwRPC_DISTMSG_REQUEST_TIMEOUT];

      if (currentTime >= expTime) {
        string project, name;

        fwRPC_msgBroker_getHandleCallee(hdl, project, name);

        fwRPC_dbg_log(
          fwRPC_DBG_EVENT_ERROR,
          "(" + project + ":" + name + ")" +
          "Core did not respond with a request status update." +
          " Attempting to reassign."
        );

        //TODO: check core state

        //attempt to reassign the call to another core. Note that no attempt is
        //made to choose a new core.
        if (fwRPC_msgBroker_getHandleCaller(hdl, project, name) == 0) {
          fwRPC_msgBroker_unregisterHandle(hdl);
          _fwRPC_broker_hdlRequest(project, name, call);
        }
      }
    }

    delay(_fwRPC_BROKER_TICK_RATE_S);
  }
}


/**
 * @internal
 *
 * Work in progress -- not for release
 *
 * Part of the fwRPC debugging interface.
 *
 * This function is executed when the registered module (using fwRPC_dbg_hook())
 * receives a trigger for a status dump.
 *
 * @reviewed 2018-06-21 @whitelisted{Callback}
 */
private void _fwRPC_broker_dbg_report() {
  dyn_string status = makeDynString(
    "nothing"
  );

  fwRPC_dbg_log(fwRPC_DBG_EVENT_DUMP, status);
}


/**
 * @internal
 *
 * Callback function for system up events. Called from broker module.
 *
 * @param project    (IN, string) project of new system
 * @param name       (IN, string) name of new system
 *
 * @returns nothing
 *
 * @throws nothing
 */
public synchronized void fwRPC_broker_newMbxCB(string project, string name) {
  if (!fwRPC_dist_isCore(name)) return;

  fwRPC_msgBroker_writeMessage(project, name, fwRPC_DISTMSG_TYPE_REGISTER);
}


/**
 * @internal
 *
 * Callback function for system down events. Called from broker module.
 *
 * @param project (IN, string) project of dropped system
 *
 * @returns nothing
 *
 * @throws nothing
 */
public synchronized void fwRPC_broker_sys_drop(string project) {
  if (!fwRPC_dist_isCore(project)) return;

  _fwRPC_broker_dropCore(project);
}


/**
 * @internal
 *
 * Remove a system from the registry. Unmaps it from service provider list.
 *
 * @param project    (IN, string) project to remove
 * @param name       (IN, string) name to remove
 *
 * @returns nothing
 *
 * @throws nothing
 */
public synchronized void _fwRPC_broker_dropCore(
  string project,
  string name=""
) {
  int droppedIndex;

  fwRPC_dbg_log(
    fwRPC_DBG_EVENT_GENERIC,
    "Core disconnected from system: " + project + ":" + name
  );

  int numCores = dynlen(g_fwRPC_broker_cores);
  for (int i = 1; droppedIndex <= numCores; ++i) {
    dyn_string coreInfo = g_fwRPC_broker_cores[i][1];

    if (coreInfo[fwRPC_DIST_NODE_PROJECT] == project &&
       (coreInfo[fwRPC_DIST_NODE_NAME]    == name || name == "")
    ) {
      droppedIndex = i;
      dynRemove(g_fwRPC_broker_cores, droppedIndex);
      break;
    }
  }

  if (droppedIndex <= g_fwRPC_broker_coreIndex) --g_fwRPC_broker_coreIndex;
}


/**
 * @internal
 *
 * Remove a client system from the registry. Does nothing.
 *
 * @param project    (IN, string) project to remove
 * @param name       (IN, string) name to remove
 *
 * @returns nothing
 *
 * @throws nothing
 */
public synchronized void _fwRPC_broker_dropClient(
  string project,
  string name=""
) {
  //do nothing
  fwRPC_dbg_log(
    fwRPC_DBG_EVENT_GENERIC,
    "Client dropped from system: " + project + ":" + name
  );
}


//------------------------------------------------------------------------------
// request handlers
//------------------------------------------------------------------------------


/**
 * @internal
 *
 * Handle disconnect message.
 *
 * @param senderProject  (IN, string)  originating project
 * @param senderName     (IN, string)  originating name
 * @param buffer         (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Callback}
 */
public void _fwRPC_broker_hdlDisconnect(
  string senderProject, string senderName,
  anytype buffer
) {
  if (fwRPC_dist_isCore(senderName)) {
    _fwRPC_broker_dropCore(senderProject, senderName);
  } else if (fwRPC_dist_isClient(senderName)) {
    _fwRPC_broker_dropClient(senderProject, senderName);
  }
}


/**
 * @internal
 *
 * Handle request message.
 *
 * @param senderProject  (IN, string)  originating project
 * @param senderName     (IN, string)  originating name
 * @param buffer         (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_broker_hdlRequest(
  string senderProject, string senderName,
  anytype buffer
) { 
  anytype hdl = buffer[fwRPC_DISTMSG_REQUEST_HDL];
  string serviceName = buffer[fwRPC_DISTMSG_REQUEST_FUNC];
  dyn_string node;

  int err = _getNextServiceProvider(serviceName, node);
  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_WARNING,
      "service not registered: " + serviceName + " client will be notified: " +
      buffer[fwRPC_DISTMSG_REQUEST_HDL]
    );

    dyn_mixed msg;
    msg[fwRPC_DISTMSG_RESULT_HDL] = hdl;
    msg[fwRPC_DISTMSG_RESULT_STATUS] = fwRPC_REQUEST_STATUS_ERROR;
    msg[fwRPC_DISTMSG_RESULT_OUTPUT] = makeError(
      getErrorCatalog(fwRPC_EX_INVALID_PARAM),
      getErrorPriority(fwRPC_EX_INVALID_PARAM),
      getErrorType(fwRPC_EX_INVALID_PARAM),
      getErrorCode(fwRPC_EX_INVALID_PARAM),
      "Called RPC function is not registered: " +
      buffer[fwRPC_DISTMSG_REQUEST_FUNC]
    );

    fwRPC_msgBroker_writeMessage(
      senderProject, senderName,
      fwRPC_DISTMSG_TYPE_SET_RESULT, msg
    );

    return;
  }

  string target_project = node[fwRPC_DIST_NODE_PROJECT];
  string target_name    = node[fwRPC_DIST_NODE_NAME];

  fwRPC_msgBroker_registerHandle(
    hdl,
    senderProject, senderName,
    target_project, target_name,
    buffer[fwRPC_DISTMSG_REQUEST_FUNC]
  );

  fwRPC_msgBroker_writeMessage(
    target_project, target_name,
    fwRPC_DISTMSG_TYPE_REQUEST, buffer
  );

  time timeout_ts = getCurrentTime() + _fwRPC_BROKER_CORE_EXPIRE_TIME_S;
  buffer[fwRPC_DISTMSG_REQUEST_TIMEOUT] = timeout_ts;
  _g_fwRPC_broker_pendingCalls[hdl] = buffer;
}


/**
 * @internal
 *
 * Handle status messages from cores.
 *
 * @param senderProject  (IN, string)  originating project
 * @param senderName     (IN, string)  originating name
 * @param buffer         (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Callback}
 */
public void _fwRPC_broker_hdlStatus(
  string senderProject, string senderName,
  anytype buffer
) {
  string id = buffer[fwRPC_DISTMSG_STATUS_HDL];

  string clientProject, clientName;
  if (fwRPC_msgBroker_getHandleCaller(id, clientProject, clientName) < 0) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR, "error retrieving caller for " + id);
    return;
  }

  fwRPC_msgBroker_writeMessage(
    clientProject, clientName,
    fwRPC_DISTMSG_TYPE_STATUS, buffer
  );

  //remove from pending list
  mappingRemove(_g_fwRPC_broker_pendingCalls, id);
}


/**
 * @internal
 *
 * Handle result messages.
 *
 * @param senderProject  (IN, string)  originating project
 * @param senderName     (IN, string)  originating name
 * @param buffer         (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Callback}
 */
public void _fwRPC_broker_hdlGetResult(
  string senderProject, string senderName,
  anytype buffer
) {
  string target_project, target_name;

  int err = fwRPC_msgBroker_getHandleCallee(buffer, target_project, target_name);

  if (!err) {
    fwRPC_msgBroker_writeMessage(
      target_project, target_name,
      fwRPC_DISTMSG_TYPE_GET_RESULT, buffer
    );
  } else {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Error retrieving callee for " + buffer + " client will be notified."
    );

    dyn_mixed msg;
    msg[fwRPC_DISTMSG_RESULT_HDL] = buffer[fwRPC_DISTMSG_RESULT_HDL];
    msg[fwRPC_DISTMSG_RESULT_STATUS] = fwRPC_REQUEST_STATUS_ERROR;
    msg[fwRPC_DISTMSG_RESULT_OUTPUT] = makeError(
      getErrorCatalog(fwRPC_EX_INVALID_PARAM),
      getErrorPriority(fwRPC_EX_INVALID_PARAM),
      getErrorType(fwRPC_EX_INVALID_PARAM),
      getErrorCode(fwRPC_EX_INVALID_PARAM),
      "Unable to retrieve result for request, hdl not registered: " +
      buffer[fwRPC_DISTMSG_RESULT_HDL]
    );

    fwRPC_msgBroker_writeMessage(
      senderProject, senderName,
      fwRPC_DISTMSG_TYPE_SET_RESULT, msg
    );
  }
}


/**
 * @internal
 *
 * Handle request set requests.
 *
 * @param senderProject (IN, string)  originating project
 * @param senderName    (IN, string)  originating name
 * @param buffer        (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Callback}
 */
public void _fwRPC_broker_hdlSetResult(
  string senderProject, string senderName,
  anytype buffer
) {
  string target_project; string target_name;

  int err = fwRPC_msgBroker_getHandleCaller(
    buffer[fwRPC_DISTMSG_RESULT_HDL],
    target_project,
    target_name
  );

  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Error retrieving caller for " + buffer[fwRPC_DISTMSG_RESULT_HDL]
    );
    return;
  }

  fwRPC_msgBroker_writeMessage(
    target_project, target_name,
    fwRPC_DISTMSG_TYPE_SET_RESULT, buffer
  );
}


/**
 * @internal
 *
 * Handle register messages.
 *
 * @param senderProject  (IN, string)  originating project
 * @param senderName     (IN, string)  originating name
 * @param buffer         (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Callback}
 */
public void _fwRPC_broker_hdlRegister(
  string senderProject, string senderName,
  anytype buffer
) {
  dyn_string coreInfo;
  string serviceName;

  coreInfo[fwRPC_DIST_NODE_PROJECT] = senderProject;
  coreInfo[fwRPC_DIST_NODE_NAME]    = senderName;

  dyn_string funcs = buffer;

  for (int i=1; i <= dynlen(g_fwRPC_broker_cores); ++i) {
    dyn_string node = g_fwRPC_broker_cores[i][1];

    if (node[fwRPC_DIST_NODE_PROJECT] == coreInfo[fwRPC_DIST_NODE_PROJECT] &&
        node[fwRPC_DIST_NODE_NAME]    == coreInfo[fwRPC_DIST_NODE_NAME]
    ) {
      for (int j=1; j<=dynlen(funcs); ++j) {
        if (dynContains(g_fwRPC_broker_cores[i][2], funcs[j])) {
          fwRPC_dbg_log(
            fwRPC_DBG_EVENT_WARNING,
            funcs[j] + " already registered for " + senderName + " @ " + 
            senderProject
          );
          continue;
        }

        //service was not registered, append to registry
        dynAppend(g_fwRPC_broker_cores[i][2], serviceName);
      }

      return;
    }
  }

  //core was not found. create a new core entry and append
  dyn_mixed entry;
  entry[1] = coreInfo;
  entry[2] = funcs;

  if (!dynlen(g_fwRPC_broker_cores)) g_fwRPC_broker_cores = makeDynMixed(entry);
  else dynAppend(g_fwRPC_broker_cores, entry);

  if (g_fwRPC_broker_coreIndex == 0) g_fwRPC_broker_coreIndex = 1;
}


/**
 * @internal
 *
 * Handle acknowledge messages.
 *
 * @param senderProject  (IN, string)  originating project
 * @param senderName     (IN, string)  originating name
 * @param buffer         (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Callback}
 */
public void _fwRPC_broker_hdlAcknowledge(
  string senderProject, string senderName,
  anytype buffer
) {
  string target_project, target_name;

  int err = fwRPC_msgBroker_getHandleCallee(buffer, target_project, target_name);
  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "request not registered, could not ack: " + buffer
    );
    return;
  }

  fwRPC_msgBroker_unregisterHandle(buffer);

  fwRPC_msgBroker_writeMessage(
    target_project, target_name,
    fwRPC_DISTMSG_TYPE_ACKNOWLEDGE, buffer
  );
}


/**
 * @internal
 *
 * Handle cancel messages.
 *
 * @param senderProject  (IN, string)  originating project
 * @param senderName     (IN, string)  originating name
 * @param buffer         (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Callback}
 */
public void _fwRPC_broker_hdlCancel(
  string senderProject, string senderName,
  anytype buffer
) {
  string target_project, target_name;
  if (fwRPC_msgBroker_getHandleCallee(buffer, target_project, target_name)) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "request not registered, unable to cancel: " + buffer
    );
    return;
  }

  fwRPC_msgBroker_writeMessage(
    target_project, target_name,
    fwRPC_DISTMSG_TYPE_CANCEL, buffer
  );
}


/**
 * @internal
 *
 * Handle recovery messages.
 *
 * @param senderProject  (IN, string)  originating project
 * @param senderName     (IN, string)  originating name
 * @param buffer         (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Callback}
 */
public void _fwRPC_broker_hdlRecover(
  string senderProject, string senderName,
  anytype buffer
) {
  int err;
  dyn_mixed calls;

  fwRPC_msgBroker_getHandlesCalledBy(
    senderProject,
    senderName,
    calls
  );

  if (dynlen(calls) == 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_GENERIC,
      "Client " + senderProject + " " + senderName +
      " requested recovery, but no active requests exist."
    );

    return;
  }

  int callsToRecover = dynlen(calls);
  for (int i = 1; i <= callsToRecover; ++i) {
    string coreProject, coreName;

    err = fwRPC_msgBroker_getHandleCallee(
      calls[i],
      coreProject, coreName
    );
    if (err < 0) {
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_ERROR,
        "unable to retrieve callee for " + calls[i]
      );
      continue;
    }

    dyn_string resp;
    resp[fwRPC_DISTMSG_RECOVER_HDL] = calls[i];
    resp[fwRPC_DISTMSG_RECOVER_CALLBACK] = buffer;

    fwRPC_msgBroker_writeMessage(
      coreProject, coreName,
      fwRPC_DISTMSG_TYPE_RECOVER, resp
    );
  }
}


/**
 * Handle service listing messages.
 *
 * @param senderProject  (IN, string)  originating project
 * @param senderName     (IN, string)  originating name
 * @param buffer         (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Callback}
 */
public void _fwRPC_broker_hdlListServices(
  string senderProject, string senderName,
  string buffer
) {
  dyn_string procedures;

  int coresToList = dynlen(g_fwRPC_broker_cores);
  for (int coreIdx = 1; coreIdx <= coresToList; ++coreIdx) {
    dyn_mixed entry = g_fwRPC_broker_cores[coreIdx];
    dyn_string funcList = entry[2];

    int functionsToList = dynlen(funcList);
    for (int procIdx = 1; procIdx <= functionsToList; ++procIdx) {
      if (!dynContains(procedures, funcList[procIdx])) {
        dynAppend(procedures, funcList[procIdx]);
      }
    }
  }

  fwRPC_msgBroker_writeMessage(
    senderProject, senderName,
    fwRPC_DISTMSG_TYPE_LISTSVC, procedures
  );
}


/**
 * @internal
 *
 * Get the next core in the queue for request processing.
 *
 *
 * Algorithm is as follows:
 *  - iterate from the current index until a core with the required service is
 *    found or the list is exhausted.
 *  - if the new index is the same as the current, the iterator is incremented
 *    by one. Otherwise, we leave the iterator as it is and go back to it for
 *    the next allocation.
 *
 * @param service (IN,  string)     the service
 * @param node    (OUT, dyn_string) the node info of the core
 *
 * @returns  0 on success
 *          -1 if a core was not found
 *
 * @throws nothing
 */
public int _getNextServiceProvider(string service, dyn_string &node) {
  if (g_fwRPC_broker_coreIndex == 0) return -1;

  int coreIdx = g_fwRPC_broker_coreIndex;

  dyn_mixed core = g_fwRPC_broker_cores[coreIdx];

  while(!dynContains(core[2], service)) {
    if (coreIdx < coreIdx) coreIdx++;
    else coreIdx = 1;

    //we've reached the end
    if (coreIdx == g_fwRPC_broker_coreIndex) {
      return -1;
    }

    core = g_fwRPC_broker_cores[coreIdx];
  }

  //if we've used the coreIdx node, just increment the index
  //otherwise we should come back to it for the next delegation
  if (coreIdx == g_fwRPC_broker_coreIndex) {
    if (coreIdx < dynlen(g_fwRPC_broker_cores)) g_fwRPC_broker_coreIndex++;
    else g_fwRPC_broker_coreIndex = 1;
  }

  node = core[1];

  return 0;
}
