/** @file
 *
 * This library contains the set of functions to request and monitor RPC calls.
 * Part of the client-side RPC library.
 *
 * @author  Victor Rodrigues (CERN, BE-ICS-SDS)
 */

#uses "CtrlSerialization"

#uses "fwRPC/fwRPC_common.ctl"
#uses "fwRPC/fwRPC_requestModel.ctl"
#uses "fwRPC/fwRPC_functionModel.ctl"
#uses "fwRPC/fwRPC_exception.ctl"
#uses "fwRPC/fwRPC_msgClient.ctl"
#uses "fwRPC/fwRPC_UUID.ctl"

#uses "fwRPC/fwRPC_debug.ctl"


/**
 * Error codes
 */

global const int fwRPC_CLIENT_E_SUCCESS = 0;
global const int fwRPC_CLIENT_E_ERROR   = -1;
global const int fwRPC_CLIENT_E_TIMEOUT = -2;

/**
 * End error codes
 */


/**
 * Collection of client subscriptions.
 * Structure:
 *  key = client context handle
 *  value =
 *    DYN_DYN_MIXED:
 *      [x]: DYN_MIXED:
 *              [1]: request handle
 *              [2]: callback function
 */
private global mapping g_fwRPC_client_subscriptions;

/**
 * Whether or not the client is running.
 * Structure:
 * key = client context handle
 * value = client with ctx is running
 */
private global mapping g_fwRPC_client_monitorIsRunning;

/**
 * Mapping of all call results.
 * Structure:
 *  key = client context handle
 *  value = mapping:
 *            key   = request handle
 *            value = request model
 */
private global mapping g_fwRPC_client_activeRequests;

/**
 *  List of available services.
 *  Structure:
 *  key = client context handle
 *  value =
 *    DYN_STRING:
 *      [x]: procedure name
 */
private global mapping g_fwRPC_client_services;


/**
 * Whether or not the client is presumed to be connected to a core or broker.
 * Structure:
 *  key = client context handle
 *  value = client with ctx is connected
 */
private global mapping g_fwRPC_client_isConnected;


/**
 * Time to wait in seconds between ticks to check for local request timeout.
 */
private global const int _fwRPC_CLIENT_TICK_RATE_S = 1;

/**
 * Time in ms to wait between poll attempts for a message response. Used for
 * getResult polls and listServices.
 */
private global const int _fwRPC_CLIENT_RESPONSE_POLL_MS = 50;


/**
 * Previous known redu states. Used when connected to a redundant system.
 * key: 
 * key: redu peer
 * value: dyn_bool:
 *          [1]: isActive
 *          [2]: isPeerAlive
 *          [3]: isRecovering
 */
private global mapping g_fwRPC_client_redu_state;


//------------------------------------------------------------------------------


/**
 * Return whether or not the client monitor is currently running.
 *
 * @returns true  if the client monitor is running
 *          false otherwise
 *
 * @throws nothing
 */
bool fwRPC_client_monitorIsRunning(anytype ctx) {
  if (!mappingHasKey(g_fwRPC_client_monitorIsRunning, ctx)) return false;
  return g_fwRPC_client_monitorIsRunning[ctx];
}


/**
 * Start the client. Required for any RPC requests, local or remote.
 * If a client with the given or generated name was already running in the
 * current CTRL context, then it is taken over.
 *
 * The client name is automatically set to the panel name, or if started in a
 * WCCOActrl manager, it will be set to a value generated by the msgClient.
 * The name can also be manually specified.
 *
 * If manually specified, the client name MUST be unique.
 *
 * @param clientName  (IN|OUT, string) unique client name.
 * @param distributed (IN, bool)       whether or not to start in distributed mode
 *
 * @returns nothing
 *
 * @throws if the connection was not successful.
 */
void fwRPC_client_startMonitor(
  anytype &ctx,
  string &clientName,
  bool distributed=false,
  bool forceReduDetect=false
) {
  mapping msgDelegates;
  int beginTick_tid = -1;

  //if we haven't been supplied a name, try to get one from the panel if we're
  //running in one.
  if (clientName == "" && isFunctionDefined("myPanelName")) {
    clientName = strsplit(myPanelName(), '.')[1];
  }

  fwRPC_dbg_hook(clientName, fwRPC_DBG_MODULE_CLIENT, "_fwRPC_client_dbg_report");

  msgDelegates[fwRPC_DISTMSG_TYPE_CONNECT]    = "_fwRPC_client_distHdlConnect";
  msgDelegates[fwRPC_DISTMSG_TYPE_STATUS]     = "_fwRPC_client_distHdlStatus";
  msgDelegates[fwRPC_DISTMSG_TYPE_SET_RESULT] = "_fwRPC_client_distHdlResult";
  msgDelegates[fwRPC_DISTMSG_TYPE_LISTSVC]    = "_fwRPC_client_distHdlListServices";

  fwRPC_msgClient_initialize(
    ctx,
    !distributed,
    fwRPC_DIST_TAG_CLIENT,
    clientName,
    msgDelegates
  );

  // initialise global state
  g_fwRPC_client_services[ctx] = makeDynString();
  mapping mapReqs;
  g_fwRPC_client_activeRequests[ctx] = mapReqs;
  dyn_dyn_mixed container;
  g_fwRPC_client_subscriptions[ctx] = container;

  g_fwRPC_client_monitorIsRunning[ctx] = true;

  int redu_status;
  reduActive(redu_status);
  bool isReduSystem = (redu_status != 0 || forceReduDetect);

  beginTick_tid = startThread("_fwRPC_client_beginTick", ctx, isReduSystem);
  if (beginTick_tid == -1) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Failed to start client tick. Future async calls will not locally time out"
    );
  }
}


private const int fwRPC_MSGCLIENT_POOL_DEFAULT_WAITING_TIME_ms = 5000;

/**
 * Similar to fwRPC_client_startMonitor() but ignores the given client name (context name).
 * Instead it assigns a context from a shared pool.
 * 
 * @param int waitForContext_ms = 0 - time for how long we want to wait for the context
 *                                   - < 0 - there is no limit how long we wait for the context
 *                                   - == 0 - no waiting for a context will be done
 *                                   - > 0 - the function will try to get a context name within a limited time
 * 
 * @return false if something has failed (i.e. we could not get a context name from the client pool)
 * 
 * @see fwRPC_client_startMonitor()
 * @reviewed 2018-08-03 @whitelisted{API}
 */
bool fwRPC_client_startMonitor_poolContext(
  anytype &ctx,
  string &clientName,
  bool distributed=false,
  bool forceReduDetect=false,
  int waitForContext_ms = fwRPC_MSGCLIENT_POOL_DEFAULT_WAITING_TIME_ms
) {
  clientName = _fwRPC_client_getContextNameFromPool(waitForContext_ms);

  if(clientName == "")
  {
    // we have failed to get a context name from the client pool 
    return false;
  }

  fwRPC_client_startMonitor(ctx, clientName, distributed, forceReduDetect);
  return true;
}


/**
 * Similar to fwRPC_client_stopMonitor() function but this one is used
 * for stopping monitor of contexts taken from shared pool.
 * This function returns the given context back to the shared pool
 * so it can be reused again during another RPC call.
 * 
 * @see fwRPC_client_stopMonitor()
 * @reviewed 2018-08-03 @whitelisted{API}
 */
void fwRPC_client_stopMonitor_poolContext(anytype ctx, string clientName) {
  fwRPC_client_stopMonitor(ctx);
  _fwRPC_client_returnContextNameToPool(clientName);
}


/**
 * This constant specifies the time we wait before trying to get
 * a context name again.
 */
private const int fwRPC_MSGCLIENT_POOL_ATTEMPT_WAITING_TIME_ms = 250;


/**
 * @param int waitForContext_ms = 0 - time for how long we want to wait for the context
 *                                   - < 0 - there is no limit how long we wait for the context
 *                                   - == 0 - no waiting for a context will be done
 *                                   - > 0 - the function will try to get a context name within a limited time
 * 
 * @return context name from the shared pool of contexts. If no context is available, empty string is returned.
 */
synchronized string _fwRPC_client_getContextNameFromPool(const int waitForContext_ms = -1)
{
  const string poolDp = fwRPC_common_getDpNameForContextPool(fwRPC_MSGCLIENT_POOL_NAME);
  const string availableContextsDpe = poolDp+".availableContexts";

  if(dpExists(availableContextsDpe) == false)
  {
    DebugTN(__FUNCTION__, "Client context pool DP does not exist. Make sure that the Core Manager was started properly.", availableContextsDpe, getStackTrace());
    return "";
  }

  const long startTime_ms = fwRPC_common_timeGetCurrentTimeInMilliSeconds();
  bool isFirstIteration = true;

  while(true)
  {
    long remainingTimeToWaitForLock_ms;
    if(waitForContext_ms < 0)
    {
      // user wants to wait forever
      remainingTimeToWaitForLock_ms = waitForContext_ms;
    }
    else if(waitForContext_ms == 0)
    {
      // user does not want to wait at all
      remainingTimeToWaitForLock_ms = 0;
    }
    else
    {
      const long newRemainingTimeToWait_ms = waitForContext_ms - (fwRPC_common_timeGetCurrentTimeInMilliSeconds()-startTime_ms);
      if(newRemainingTimeToWait_ms < 0 && isFirstIteration == false)
      {
        // We tried to get the context name at least once already and now we have run out of time
        return "";
      }
      remainingTimeToWaitForLock_ms = newRemainingTimeToWait_ms;
    }


    const bool isLocked = fwRPC_common_dpLock(availableContextsDpe, remainingTimeToWaitForLock_ms);

    if(isLocked == false)
    {
      // We were not able to get the lock for the pool
      return ""; // Therefore we return "" to indicate a problem
    }

    dyn_string availableContexts;
    dpGet(availableContextsDpe, availableContexts);

    if(dynlen(availableContexts) > 0)
    {
      const string contextName = availableContexts[1];
      dynRemove(availableContexts, 1);
      
      const int errFlag = dpSetWait(availableContextsDpe, availableContexts);
      const dyn_errClass lastErr = getLastError();

      fwRPC_common_dpUnlock(availableContextsDpe);

      if(errFlag < 0 || dynlen(lastErr) > 0)
      {
        DebugTN(__FUNCTION__, "A problem with dpSetWait() call detected in fwRPC.", availableContextsDpe, errFlag, lastErr, getStackTrace());
        return "";
      }

      return contextName;
    }
    else
    {
      fwRPC_common_dpUnlock(availableContextsDpe);

      /**
       * WARNING: know bug, @see ENS-22030
       * The problem is that when the context pool is empty and we put back one context,
       * it is immediately reused by some other client. And we know that reusing the same
       * context too frequently is a BAD IDEA (ENS-22030). Unfortunately with the current
       * fwRPC architecture, we cannot simply avoid this problem.
       * 
       * RECOMMENDATION: Make sure that the pool is big enough so it is never depleted.
       * If the pool is empty, the following code will work fine and it will wait for another context
       * but some other parts of fwRPC may fail because of the frequent reusing of the same contexts.
       */
      
      /**
       * There are no contexts in the pool. So we have to wait for somebody to return
       * a context.
       */
      long remainingTimeToWaitForPoolChange_ms;
      if(waitForContext_ms < 0)
      {
        // user wants to wait forever
        remainingTimeToWaitForPoolChange_ms = waitForContext_ms;
      }
      else if(waitForContext_ms == 0)
      {
        // user does not want to wait at all
        return "";
      }
      else
      {
        const long newRemainingTimeToWait_ms = waitForContext_ms - (fwRPC_common_timeGetCurrentTimeInMilliSeconds()-startTime_ms);
        if(newRemainingTimeToWait_ms <= 0)
        {
          // We have run out of time
          return "";
        }
        remainingTimeToWaitForPoolChange_ms = newRemainingTimeToWait_ms;
      }

      time waitForContext_timeout;
      setPeriod(waitForContext_timeout, 0, remainingTimeToWaitForPoolChange_ms);

      dyn_anytype returnValues;
      bool timerExpired;
      dpWaitForValue(makeDynString(availableContextsDpe),
                                    makeDynAnytype(),
                                    makeDynString(availableContextsDpe),
                                    returnValues, waitForContext_timeout, timerExpired);

      if(timerExpired == true)
      {
        return ""; // we have run out of time
      }
    }

    isFirstIteration = false;
  }

  DebugTN(__FUNCTION__, "fwRPC algorithm to get context from clinet pool has FAILED! We have executed code that should never be executed (in theory)", getStackTrace());
  return "";
}


private const string fwRPC_MSGCLIENT_POOL_NAME = "clientPool";

/**
 * Puts the given context back to the shared pool
 * 
 * @param string contextName - name of the context that should be put back to the shared pool
 * @return true if the operation was successful, else false.
 */
synchronized bool _fwRPC_client_returnContextNameToPool(const string contextName)
{
  const string poolDp = fwRPC_common_getDpNameForContextPool(fwRPC_MSGCLIENT_POOL_NAME);
  const string availableContextsDpe = poolDp+".availableContexts";

  if(dpExists(availableContextsDpe) == false)
  {
    DebugTN(__FUNCTION__, "Client context pool DP does not exist. Make sure that the Core Manager was started properly.", availableContextsDpe, getStackTrace());
    return false;
  }

  const bool isLocked = fwRPC_common_dpLock(availableContextsDpe);

  if(isLocked == true)
  {

    dyn_string availablContexts;
    dpGet(availableContextsDpe, availablContexts);
    
    /**
     * We put the context in the pool, AT THE END of the list.
     * It would be more effecient to put it at the beginning,
     * however there seems to be a problem with Ring Buffer indexes
     * when the context are used too frequently.
     * @see PSEN-2287, ENS-22030
     */
    dynAppend(availablContexts, contextName);
    const int errFlag = dpSetWait(availableContextsDpe, availablContexts);
    const dyn_errClass lastErr = getLastError();

    fwRPC_common_dpUnlock(availableContextsDpe);
    if(errFlag < 0 || dynlen(lastErr) > 0)
    {
      DebugTN(__FUNCTION__, "A problem with dpSetWait() call detected in fwRPC.", availableContextsDpe, errFlag, lastErr);
      return false;
    }

    return true;
  }
  else
  {
    return false;
  }
}


/**
 * Ensures that shared context pool for clients has a minimal size.
 * 
 * @parem uint minPoolSize - min size of the client context pool
 */
void fwRPC_client_ensureMinimumPoolSize(const uint minPoolSize)
{
  fwRPC_common_ensureMinimumPoolSize(minPoolSize, fwRPC_MSGCLIENT_POOL_NAME);
}


/**
 * Stop the current client monitor. Detaches any message listeners in local or
 *   distributed mode.
 *
 * @returns nothing
 *
 * @throws if the monitor is not running
 *         if dpQueryConnect failed
 */
void fwRPC_client_stopMonitor(anytype ctx) {
  _fwRPC_client_assertRunning(ctx, true);

  int err = fwRPC_msgClient_stop(ctx);

  g_fwRPC_client_monitorIsRunning[ctx] = false;
}


/**
 * @internal
 *
 * Work in progress -- not for release
 *
 * Part of the fwRPC debugging interface.
 *
 * This function is executed when the registered module (using fwRPC_dbg_hook())
 * receives a trigger for a status dump.
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-22 @whitelisted{Callback}
 */
private void _fwRPC_client_dbg_report(anytype ctx) {
  dyn_string state = makeDynString(
    fwRPC_common_cleanProjectName(getSystemName()),
    fwRPC_msgClient_getName(ctx)
  );

  fwRPC_dbg_log(fwRPC_DBG_EVENT_DUMP, state);
}


/** @internal
 *
 * Performs a check on each tick for local request timeouts. Note that the
 * timeout value is ignored once status updates are received from the server.
 * Local timeout is meant for the case in which a server is non-responsive.
 *
 * In addition, beginTick will also initialise and tear-down the monitoring of
 * changes to the redundant system if running in one.
 *
 * @param isRedundantSystem (IN, bool) whether the client is connected to a
 * redundant system
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-22 @whitelisted{Thread}
 */
private void _fwRPC_client_beginTick(
  anytype ctx,
  bool isRedundantSystem = false
) {
  time currentTime; 
  dyn_mixed request;
  mixed expTime;

  if (isRedundantSystem) {
    _fwRPC_client_linkRedu();
  }

  while (true) {
    if (!g_fwRPC_client_monitorIsRunning[ctx] && isRedundantSystem) {
      _fwRPC_client_unlinkRedu();
      return;
    }

    currentTime = getCurrentTime();

    int len = mappinglen(g_fwRPC_client_activeRequests[ctx]);
    for (int reqIdx = 1; reqIdx <= len; ++reqIdx) {
      request = mappingGetValue(g_fwRPC_client_activeRequests[ctx], reqIdx);
      expTime = request[fwRPC_REQUEST_MODEL_EXPIRETIME];
      if (expTime == -1) continue;

      if (currentTime >= expTime) {
        _fwRPC_client_processStatusUpdate(
          ctx,
          request[fwRPC_REQUEST_MODEL_ID],
          fwRPC_REQUEST_STATUS_TIMEOUT
        );
      }
    }

    delay(_fwRPC_CLIENT_TICK_RATE_S);
  }
}


//------------------------------------------------------------------------------
// START CLIENT REQUEST API
//------------------------------------------------------------------------------


/**
 * Attempt to recover previous unacknowledged requests. Useful in case of client
 *   crash.
 *
 * The callback function should follow the same signature as the subscription
 *   callback function.
 *
 * @param callback (IN, string) callback function
 *
 * @returns fwRPC_CLIENT_E_SUCCESS  on success
 *          fwRPC_CLIENT_E_ERROR    otherwise
 *
 * @throws if the monitor is not running
 */
int fwRPC_client_recover(anytype ctx, string callback) {
  _fwRPC_client_assertRunning(ctx, true);

  int err = fwRPC_msgClient_sendRequest(ctx,fwRPC_DISTMSG_TYPE_RECOVER, callback);

  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Error sending recovery message: " + err
    );
    return fwRPC_CLIENT_E_ERROR;
  }

  return fwRPC_CLIENT_E_SUCCESS;
}


/**
 * Attempt to acknowledge a previous RPC request.
 * Acknowledging a requests will cause the rpcHandle to no longer be valid for
 *   the current request instance.
 *
 * @param rpcHandle (IN, anytype) RPC request handle
 *
 * @returns fwRPC_CLIENT_E_SUCCESS  on success
 *          fwRPC_CLIENT_E_ERROR    otherwise
 *
 * @throws if the client was not initialized
 *         if the monitor is not running
 */
int fwRPC_client_acknowledge(anytype ctx, anytype rpcHandle) {
  _fwRPC_client_assertRunning(ctx, true);
  _fwRPC_client_assertValidHdl(rpcHandle);

  _fwRPC_client_unsubscribeAllCallbacks(ctx, rpcHandle);

  if (mappingHasKey(g_fwRPC_client_activeRequests[ctx], rpcHandle)) {
    mappingRemove(g_fwRPC_client_activeRequests[ctx], rpcHandle);
  }

  int err = fwRPC_msgClient_sendRequest(
    ctx,
    fwRPC_DISTMSG_TYPE_ACKNOWLEDGE,
    rpcHandle
  );

  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Error sending acknowledge message: " + err
    );
  }
  
  return err < 0 ? fwRPC_CLIENT_E_ERROR : fwRPC_CLIENT_E_SUCCESS;
}


/**
 * Retrieve a result from a previous RPC request. If timeout is > 0, then it
 * will wait (blocking) until the request is complete.
 *
 * If the return value is not fwRPC_REQUEST_STATUS_DONE or
 * fwRPC_REQUEST_STATUS_ERROR, then the result parameter will be undefined.
 *
 * @param rpcHandle   (IN,  anytype) RPC request handle
 * @param result      (OUT, mixed)   the result, if timeout not reached
 * @param acknowledge (IN,  bool)    whether or not to automatically acknowledge 
 *                                   the request on success
 * @param timeout     (IN, int)      timeout, in MS to wait for a response
 *
 * @returns  result code            on success
 *           last known status      if result is not ready
 *           fwRPC_CLIENT_E_ERROR   on failure
 *           fwRPC_CLIENT_E_TIMEOUT on result retrieval timeout
 *
 * @throws if the monitor is not running
 *         if the provided handle is invalid
 */
synchronized int fwRPC_client_getResult(
  anytype ctx,
  anytype rpcHandle,
  mixed &result,
  bool acknowledge = true,
  int timeout = 5000
) {
  _fwRPC_client_assertRunning(ctx, true);
  _fwRPC_client_assertValidHdl(rpcHandle);

  int lastStatus = -1;
  int totalMs = 0;

  while (
    lastStatus != fwRPC_REQUEST_STATUS_DONE &&
    lastStatus != fwRPC_REQUEST_STATUS_ERROR &&
    totalMs <= timeout
  ) {
    lastStatus = g_fwRPC_client_activeRequests[ctx][rpcHandle][fwRPC_REQUEST_MODEL_STATUS];

    delay(0, _fwRPC_CLIENT_RESPONSE_POLL_MS);
    totalMs += _fwRPC_CLIENT_RESPONSE_POLL_MS;
  }

  if (totalMs >= timeout) {
    if (lastStatus == fwRPC_REQUEST_STATUS_NONE) return fwRPC_CLIENT_E_TIMEOUT;
    return lastStatus;
  }

  //Note that it is illegal to return a result with an empty string so the
  //following is a valid check
  if (g_fwRPC_client_activeRequests[ctx][rpcHandle][fwRPC_REQUEST_MODEL_RETVAL] == fwRPC_REQUEST_INVALID_RETVAL) {
    int err;
    mixed _result;

    err = fwRPC_msgClient_sendRequest(
      ctx,
      fwRPC_DISTMSG_TYPE_GET_RESULT, rpcHandle
    );

    if (err < 0) {
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_ERROR,
        "Error sending request message"
      );

      return fwRPC_CLIENT_E_ERROR;
    }

    while (totalMs <= timeout) {
      _result = g_fwRPC_client_activeRequests[ctx][rpcHandle][fwRPC_REQUEST_MODEL_RETVAL];
      if (_result != fwRPC_REQUEST_INVALID_RETVAL) {
        result = _result;
        break;
      }

      delay(0, _fwRPC_CLIENT_RESPONSE_POLL_MS);
      totalMs += _fwRPC_CLIENT_RESPONSE_POLL_MS;
    }

    if (totalMs >= timeout) return fwRPC_CLIENT_E_TIMEOUT;
  } else {
    result = g_fwRPC_client_activeRequests[ctx][rpcHandle][fwRPC_REQUEST_MODEL_RETVAL];
  }

  if (acknowledge) fwRPC_client_acknowledge(ctx, rpcHandle);
  return lastStatus;
}


/**
 * Attempt an asynchronous RPC call to a registered function with given 
 *   parameters.
 * Returns a request handle which is used to reference the particular requests
 *   when using other functions such as for subscription.
 *
 * Note that it there is no guarantee as to which system the function is called
 *   at when in distributed mode, nor is it consistent, as requests are balanced
 *   by the broker.
 *
 * A timeout value can be specified which, if set to a positive value, the
 * client monitor will locally timeout the request if no response has been
 * received by the broker by that time. Timeout is specified in milliseconds.
 *
 * @param rpcHandle (OUT, anytype) RPC request handle
 * @param func      (IN,  string)  function to call
 * @param params    (IN,  anytype) function parameters
 * @param timeout   (IN,  int)     time to wait before setting state to TIMEOUT
 *
 * @returns fwRPC_CLIENT_E_SUCCESS on success
 *          fwRPC_CLIENT_E_ERROR   otherwise
 *
 * @throws if the monitor is not running
 *         if function is an empty string
 */
int fwRPC_client_call(
  anytype ctx,
  anytype &rpcHandle,
  string  func,
  anytype params,
  bool    autoAck=false,
  int     timeout=0
) {
  _fwRPC_client_assertRunning(ctx, true);

  if (func == "") {
    fwException_throw(fwRPC_EX_INVALID_PARAM, "func can not be empty.");
    return fwRPC_CLIENT_E_ERROR;
  }

  dyn_mixed request;
  dyn_mixed msg;
  anytype func_timeout;
  int err;

  rpcHandle = fwRPC_UUID_generateUUID4();
  request = fwRPC_requestModel_new(
    rpcHandle,
    params,
    func,
    fwRPC_REQUEST_STATUS_NONE
  );

  msg[fwRPC_DISTMSG_REQUEST_HDL]    = rpcHandle;
  msg[fwRPC_DISTMSG_REQUEST_FUNC]   = func;
  msg[fwRPC_DISTMSG_REQUEST_PARAMS] = params;

  //set exp time
  if (timeout <= 0) func_timeout = fwRPC_FUNCTION_MODEL_NO_TIMEOUT;
  else func_timeout = (time) (getCurrentTime() + timeout);
  request[fwRPC_REQUEST_MODEL_EXPIRETIME] = func_timeout;

  request[fwRPC_REQUEST_MODEL_AUTOACK] = autoAck;

  g_fwRPC_client_activeRequests[ctx][rpcHandle] = request;

  err = fwRPC_msgClient_sendRequest(ctx, fwRPC_DISTMSG_TYPE_REQUEST, msg);

  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Error seding client call request: " + err
    );

    return fwRPC_CLIENT_E_ERROR;
  }

  return fwRPC_CLIENT_E_SUCCESS;
}


/**
 * Attempt a synchronous RPC call to a registered function with given parameters.
 * Will wait for a result (blocking) until timeout is reached.
 *
 * This function has the same effect as calling fwRPC_client_call() followed by
 *   fwRPC_client_getResult().
 *
 * @param func        (IN,  string)   the name of the function to call
 * @param params      (IN,  anytype)  parameters ot pass to the function
 * @param result      (OUT, mixed)   result of the function call
 * @param acknowledge (IN,  bool)     wheter or not to automatically acknowledge on success
 * @param timeout     (IN,  int)      timeout, in MS to wait for a response
 *
 * @return last known result status
 *         fwRPC_CLIENT_E_ERROR on error
 *         fwRPC_CLIENT_E_TIMEOUT on timeout
 *
 * @throws if the monitor is not running
 *         if func is an empty string
 */
int fwRPC_client_callSync(
  anytype ctx,
  string func,
  anytype params,
  mixed &result,
  bool acknowledge=true,
  int timeout=5000
) {
  anytype rpcHandle;

  int err = fwRPC_client_call(ctx, rpcHandle, func, params);
  if (err != fwRPC_CLIENT_E_SUCCESS) return err;

  return fwRPC_client_getResult(ctx, rpcHandle, result, acknowledge, timeout);
}


/**
 * Attempt to cancel an active request.
 * Note that it is not guaranteed that a cancellation will take place before the
 *   request completes.
 * 
 * @param rpcHandle (IN, anytype) RPC request handle
 *
 * @returns fwRPC_CLIENT_E_SUCCESS on success
 *          fwRPC_CLEINT_E_ERROR   otherwise
 *
 * @throws if the monitor is not running
 */
int fwRPC_client_cancel(anytype ctx, anytype rpcHandle) {
  _fwRPC_client_assertRunning(ctx, true);
  _fwRPC_client_assertValidHdl(rpcHandle);

  int err = fwRPC_msgClient_sendRequest(ctx, fwRPC_DISTMSG_TYPE_CANCEL, rpcHandle);
  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Error sending cancel request: " + err
    );
    return fwRPC_CLIENT_E_ERROR;
  }

  return fwRPC_CLIENT_E_SUCCESS;
}


/**
 * Retrieve a list of active services.
 * Caches the list of services until this function is called again with refresh
 *   set to true.
 * If the client is in local mode, then the timeout parameter may be ignored.
 *
 * @param services (OUT, dyn_string) list of available services
 * @param refresh  (IN,  bool) whether or not to refresh the list of services
 * @param timeout  (IN,  int)  in distributed mode, the amount of time to wait
 *                             in MS for a response.
 *
 * @returns fwRPC_CLIENT_E_SUCCESS on success,
 *          fwRPC_CLIENT_E_TIMEOUT on timeout
 *          fwRPC_CLIENT_E_ERROR   otherwise
 *
 * @throws if the client COMM was not initialized when in distributed mode.
 *
 * @reviewed 2018-06-22 @whitelisted{Used in Unit Tests}
 */
int fwRPC_client_listServices(
  anytype ctx,
  dyn_string &services,
  bool refresh=false,
  int timeout=5000
) {
  _fwRPC_client_assertRunning(ctx, true);

  if (refresh) {
    int err;

    dynClear(g_fwRPC_client_services[ctx]);

    err = fwRPC_msgClient_sendRequest(ctx,fwRPC_DISTMSG_TYPE_LISTSVC);
    if (err < 0) {
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_ERROR,
        "Error sending client listsvc request: " + err
      );
      return fwRPC_CLIENT_E_ERROR;
    }

    while(!dynlen(g_fwRPC_client_services[ctx]) && timeout > 0) {
      delay(0, _fwRPC_CLIENT_RESPONSE_POLL_MS);
      timeout -= _fwRPC_CLIENT_RESPONSE_POLL_MS;
    }

    if (timeout <= 0) return fwRPC_CLIENT_E_TIMEOUT;
  }

  services = g_fwRPC_client_services[ctx];
  return fwRPC_CLIENT_E_SUCCESS;
}


/**
 * Retrieve the status of a RPC request.
 *
 * @param rpcHandle (IN, anytype)  RPC request handle
 *
 * @returns last known status for the given request
 *
 * @throws if the monitor is not running
 *
 * @reviewed 2018-06-22 @whitelisted{Used in Unit Tests}
 */
int fwRPC_client_getStatus(anytype ctx, anytype rpcHandle) {
  _fwRPC_client_assertRunning(ctx, true);
  _fwRPC_client_assertValidHdl(rpcHandle);

  if (!mappingHasKey(g_fwRPC_client_activeRequests[ctx], rpcHandle)) {
    return fwRPC_REQUEST_STATUS_NONE;
  }

  return g_fwRPC_client_activeRequests[ctx][rpcHandle][fwRPC_REQUEST_MODEL_STATUS];
}


/**
 * Subscribe a callback function to the status change event of an async RPC call
 *
 * Callback function must have the following signatue:
 *  void callback(anytype callHandle, int status)
 *
 * The callHandle parameter passed to the callback function is the same as the
 *  rpcRef passed to this function.
 *
 * @param rpcRef   (IN, anytype) async RPC call handle
 * @param callback (IN, string)  name of the function to call on event.
 *
 * @returns nothing
 *
 * @throws if the monitor is not running
 *         if rpcRef is not a valid handle
 */
void fwRPC_client_subscribeCallStatus(anytype ctx, anytype rpcRef, string callback) {
  _fwRPC_client_assertRunning(ctx, true);
  _fwRPC_client_assertValidHdl(rpcRef);

  if (callback == "") {
    fwException_throw(
      fwRPC_EX_INVALID_PARAM, "callback can not be empty"
    );

    return;
  }

  int status;
  dyn_mixed pair;

  pair = makeDynMixed(rpcRef, callback);
  synchronized(g_fwRPC_client_subscriptions) {
    dynAppend(g_fwRPC_client_subscriptions[ctx], pair);
  }

  if (!mappingHasKey(g_fwRPC_client_activeRequests[ctx], rpcRef)) return;

  //check if we already have some state data, and trigger db if so
  status = g_fwRPC_client_activeRequests[ctx][rpcRef][fwRPC_REQUEST_MODEL_STATUS];
  if (status != fwRPC_REQUEST_STATUS_NONE) {
    _fwRPC_client_executeCB(
      callback,
      rpcRef,
      status
    );
  }
}


/**
 * Unsubscribe a callback function from the status change event of an async RPC
 *  call.
 *
 * The specified callback function and call handle must be the same pair as a
 *  currently subscribed pair.
 *
 * @param rpcRef   (IN, anytype) async RPC call handle
 * @param callback (IN, string)  name of the function to call on event.
 *
 * @returns nothing
 *
 * @throws if the monitor is not running
 *         if rpcRef is not a valid handle
 *
 * @reviewed 2018-06-22 @whitelisted{Used in Unit Tests}
 */
void fwRPC_client_unsubscribeCallStatus(
  anytype ctx,
  anytype rpcRef,
  string callback
) {
  _fwRPC_client_assertRunning(ctx, true);
  _fwRPC_client_assertValidHdl(rpcRef);

  if (callback == "") {
    fwException_throw(
      fwRPC_EX_INVALID_PARAM, "callback can not be empty"
    );

    return;
  }

  synchronized(g_fwRPC_client_subscriptions) {
    int numSubs = dynlen(g_fwRPC_client_subscriptions[ctx]);
    for (int subIdx = 1; subIdx <= numSubs; ++subIdx) {
      if (g_fwRPC_client_subscriptions[ctx][subIdx][1] == rpcRef &&
          g_fwRPC_client_subscriptions[ctx][subIdx][2] == callback
      ) {
        dynRemove(g_fwRPC_client_subscriptions[ctx], subIdx);
        break;
      }
    }
  }
}



/**
 * Check whether or not the client is connected to a core / broker.
 *
 * @returns true  if connected
 *          false otherwise
 *
 * @throws nothing
 */
bool fwRPC_client_isConnected(anytype ctx) {
  return g_fwRPC_client_isConnected[ctx];
}


//------------------------------------------------------------------------------
// END CLIENT REQUEST API
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// START DISTRIBUTED MODE MSG HANDLING
//------------------------------------------------------------------------------


/**
 * @internal
 *
 * msgClient delegate.
 *
 * Handle client connection messages. Received when a core / broker connects to
 * this particular client.
 *
 * @param project (IN, string)  source project
 * @param name    (IN, string)  source name
 * @param msg     (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-22 @whitelisted{Callback}
 */
private void _fwRPC_client_distHdlConnect(
  anytype ctx,
  string project, 
  string name,
  anytype msg
) {
  g_fwRPC_client_isConnected[ctx] = true;

  fwRPC_dbg_log(
    fwRPC_DBG_EVENT_GENERIC,
    "Client connected to core.."
  );
}


/**
 * @internal
 *
 * msgClient delegate.
 *
 * Handle result messages. Stores the result of an async RPC call for future
 *   retrieval.
 *
 * @param project (IN, string)  source project
 * @param name    (IN, string)  source name
 * @param msg     (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Callback}
 */
private void _fwRPC_client_distHdlResult(
  anytype ctx,
  string project,
  string name,
  anytype msg
) {
  _fwRPC_client_processStatusUpdate(
    ctx,
    msg[fwRPC_DISTMSG_RESULT_HDL],
    msg[fwRPC_DISTMSG_RESULT_STATUS],
    msg[fwRPC_DISTMSG_RESULT_OUTPUT]
  );
}


/**
 * @internal
 *
 * msgClient delegate.
 *
 * Handle service listing messages. Stores available services for future use.
 *
 * @param project (IN, string)  source project
 * @param name    (IN, string)  source name
 * @param msg     (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-22 @whitelisted{Callback}
 */
private void _fwRPC_client_distHdlListServices(
  anytype ctx,
  string project,
  string name,
  anytype msg
) {
  g_fwRPC_client_services[ctx] = msg;
}


/**
 * @internal
 *
 * msgClient delegate.
 *
 * Handle request status updates from a core. Updates the local status registry
 * for the given request.
 *
 * If the status update was sent as a response to a recovery message from the
 * client, then it may contain a callback function. This function will be
 * subscribed to the given request if so.
 *
 * @param project (IN, string)  source project
 * @param name    (IN, string)  source name
 * @param msg     (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 *
 * @reviewed 2018-06-21 @whitelisted{Callback}
 */
private void _fwRPC_client_distHdlStatus(
  anytype ctx,
  string project,
  string name,
  anytype msg
) {
  string hdl = msg[fwRPC_DISTMSG_STATUS_HDL];
  int status = msg[fwRPC_DISTMSG_STATUS_STATUS];

  if (dynlen(msg) >= fwRPC_DISTMSG_STATUS_CALLBACK) {
    fwRPC_client_subscribeCallStatus(ctx, hdl, msg[fwRPC_DISTMSG_STATUS_CALLBACK]);
  }

  _fwRPC_client_processStatusUpdate(ctx, hdl, status);
}


//------------------------------------------------------------------------------
// END DISTRIBUTED MODE MSG HANDLING
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// PRIVATE UTILS
//------------------------------------------------------------------------------


/**
 * @internal
 *
 * Unsubscribes all client-side status callbacks from a given request.
 *
 * @param rpcRef (IN, anytype) async RPC call handle
 *
 * @returns nothing
 *
 * @throws if the monitor is not running
 */
private void _fwRPC_client_unsubscribeAllCallbacks(anytype ctx, anytype rpcRef) {
  _fwRPC_client_assertRunning(ctx, true);

  synchronized(g_fwRPC_client_subscriptions) {
    for (int i = 1; i <= dynlen(g_fwRPC_client_subscriptions[ctx]); ++i) {
      if (g_fwRPC_client_subscriptions[ctx][i][1] == rpcRef) {
        dynRemove(g_fwRPC_client_subscriptions[ctx], i);
        --i;
      }
    }
  }
}


/**
 * @internal
 *
 * Attempts to execute the client status callback.
 *
 * @param callback (IN, string)  function to execute
 * @param hdl      (IN, anytype) hdl to pass to client CB
 * @param status   (IN, int)     status to pass to client CB
 *
 * @returns nothing
 *
 * @throws if the callback could not be executed
 */
private void _fwRPC_client_executeCB(string callback, anytype hdl, int status) {
  int tid = startThread(
    callback,
    hdl,
    status
  );

  if (tid == -1) {
    fwException_throw(
      fwRPC_EX_CLIENT_CALLBACK,
      "Could not execute client callback for status event: " + callback
    );
  }
}


/**
 * @internal
 *
 * Process a request status update. Updates all local request models and calls
 * user update CBs.
 *
 * @param rpcHandle (IN, anytype) request handle
 * @param status    (IN, int)     new status
 * @param retval    (IN, mixed)   result value to update request with, if
 *                                applicable
 *
 * @returns nothing
 *
 * @throws nothing
 */
private void _fwRPC_client_processStatusUpdate(
  anytype ctx,
  anytype rpcHandle,
  int status,
  mixed retval = fwRPC_REQUEST_INVALID_RETVAL
) {
  dyn_mixed request;
  bool skipStatus = false;

  if (!mappingHasKey(g_fwRPC_client_activeRequests[ctx], rpcHandle)) {
    request = fwRPC_requestModel_new(rpcHandle);
  } else {
    request = g_fwRPC_client_activeRequests[ctx][rpcHandle];
    if (request[fwRPC_REQUEST_MODEL_STATUS] == status) skipStatus = true;
  }

  request[fwRPC_REQUEST_MODEL_EXPIRETIME] = -1;

  request[fwRPC_REQUEST_MODEL_STATUS] = status;
  if (retval != fwRPC_REQUEST_INVALID_RETVAL) request[fwRPC_REQUEST_MODEL_RETVAL] = retval;
  g_fwRPC_client_activeRequests[ctx][rpcHandle] = request;

  if (!skipStatus) {
    synchronized(g_fwRPC_client_subscriptions) {
      int numSubs = dynlen(g_fwRPC_client_subscriptions[ctx]);
      for (int subIdx = 1; subIdx <= numSubs; ++subIdx) {
        if (g_fwRPC_client_subscriptions[ctx][subIdx][1] == rpcHandle) {
          _fwRPC_client_executeCB(
            g_fwRPC_client_subscriptions[ctx][subIdx][2],
            rpcHandle,
            status
          );
        }
      }
    }
  }

  if (
    fwRPC_requestModel_hasTerminated(request) && 
    request[fwRPC_REQUEST_MODEL_AUTOACK]
  ) {
    fwRPC_client_acknowledge(ctx, rpcHandle);
  }
}


/**
 * @internal
 *
 * Assert that the client is in the specified running state. Throws an error
 * if the client is not running.
 * 
 * @param running (IN, bool) state to assert against
 *
 * @returns nothing
 *
 * @throws if the running state does not match
 */
private void _fwRPC_client_assertRunning(anytype ctx, bool running) {
  if (!mappingHasKey(g_fwRPC_client_monitorIsRunning, ctx) && running) {
    fwException_throw(
      fwRPC_EX_CLIENT_MONITOR,
      "Monitor should " + (running ? "" : "not") + " be running."
    );
  }

  if (g_fwRPC_client_monitorIsRunning[ctx] != running) {
    fwException_throw(
      fwRPC_EX_CLIENT_MONITOR,
      "Monitor should " + (running ? "" : "not") + " be running."
    );
  }

  if (fwRPC_msgClient_isRunning(ctx) != running) {
    fwException_throw(
      fwRPC_EX_CLIENT_MONITOR,
      "Client is " + (g_fwRPC_client_monitorIsRunning[ctx] ? "" : "not") + " running" +
      " but msgClient is " + (fwRPC_msgClient_isRunning(ctx) ? "" : "not") + " running."
    );
  }
}


/**
 * @internal
 *
 * Assert that the handle is a valid rpc handle.
 *
 * @param handle (IN, anytype) handle to check
 * 
 * @throws if !fwRPC_common_isValidHdl(handle)
 */
private void _fwRPC_client_assertValidHdl(anytype handle) {
  if (!fwRPC_common_isValidHdl(handle)) {
    fwException_throw(
      fwRPC_EX_INVALID_PARAM, "Invalid RPC handle: " + handle
    );
  }
}


//------------------------------------------------------------------------------
// BEGIN REDUNDANT SYSTEM MONITOR
//------------------------------------------------------------------------------


/**
 * @internal
 * 
 * Set up redudnant system monitoring. Connects to redundant system info DPs.
 *
 * @returns nothing
 *
 * @throw nothing
 */
private void _fwRPC_client_linkRedu() {
  const string fwRPC_REDU_STATUS_ACTIVE = ".Status.Active"   + fwRPC_DPE_VALUE;
  const string fwRPC_REDU_PEER_ALIVE    = ".PeerAlive.Link0" + fwRPC_DPE_VALUE;
  const string fwRPC_REDU_IS_RECOVERING = ".IsRecovering"    + fwRPC_DPE_VALUE;
  const string query = "SELECT '" +
                          fwRPC_REDU_STATUS_ACTIVE + "', '" +
                          fwRPC_REDU_PEER_ALIVE    + "', '" +
                          fwRPC_REDU_IS_RECOVERING + "' "   +
                        "FROM '*' " +
                        "WHERE _DPT =\"_ReduManager\"";

  int err = dpQueryConnectSingle(
    "_fwRPC_client_reduEventCB",
    false,
    "",
    query
  );
  
  if (err == -1) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
      "redu monitor: could not connect query: " + query
    );
  }
}


/**
 * @internal
 *
 * Disconnect from redundant system info DPs.
 *
 * @returns nothing
 *
 * @throws nothing
 */
private void _fwRPC_client_unlinkRedu() {
  int err = dpQueryDisconnect("_fwRPC_client_reduEventCB", "");
  if (err == -1) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
      "redu monitor: could not disconnect query"
    );
  }
}


/**
 * @internal
 *
 * Handle some change in redu settings. Attempt to re request previously
 * incomplete calls if a peer switch has occurred.
 *
 * @param _       (IN, anytype)         ignored
 * @param results (IN, dyn_dyn_anytype) dpQueryConnect results
 *
 * @returns nothing
 *
 * @throws nothing
 */
private void _fwRPC_client_reduEventCB(anytype _, dyn_dyn_anytype results)
synchronized(g_fwRPC_client_redu_state) {
  fwRPC_dbg_log(fwRPC_DBG_EVENT_WARNING, "REDU EVENT DETECTED, processing.");

  bool doRecall = false;

  //NOTE TODO XXX -----
  //This might fail, not tested with multiple clients sharing the same
  //context.....!
  //NOTE TODO XXX -----

  for (int i=2; i<=dynlen(results); ++i) {
    string mngr         = results[i][1];
    bool   isActive     = results[i][2];
    bool   isPeerAlive  = results[i][3];
    bool   isRecovering = results[i][4];

    if (mappingHasKey(g_fwRPC_client_redu_state, mngr)) {
      if (isActive && g_fwRPC_client_redu_state[mngr][1] != isActive) {

      }

      if (!isRecovering && isActive) doRecall = true;
    }

    g_fwRPC_client_redu_state[mngr] = makeDynBool(
      isActive,
      isPeerAlive,
      isRecovering
    );
  }

  if (doRecall) {
    dyn_anytype ctxs = mappingKeys(g_fwRPC_client_activeRequests);
    for (int ctx_idx=1; ctx_idx <= dynlen(ctxs); ++ctx_idx) {
      anytype ctx = ctxs[ctx_idx];
      int numCalls = mappinglen(g_fwRPC_client_activeRequests[ctx]);

      for (int cidx=1; cidx<=numCalls; ++cidx) {
        dyn_mixed request = mappingGetValue(g_fwRPC_client_activeRequests[ctx], cidx);

        if (!fwRPC_requestModel_hasTerminated(request)) {
          anytype hdl = request[fwRPC_REQUEST_MODEL_ID];
          request[fwRPC_REQUEST_MODEL_STATUS] = fwRPC_REQUEST_STATUS_NONE;

          //we lost timeout data...
          request[fwRPC_REQUEST_MODEL_EXPIRETIME] = -1;

          g_fwRPC_client_activeRequests[ctx][hdl] = request;

          dyn_mixed msg;
          msg[fwRPC_DISTMSG_REQUEST_HDL] = hdl;
          msg[fwRPC_DISTMSG_REQUEST_FUNC] = request[fwRPC_REQUEST_MODEL_FUNCTION];
          msg[fwRPC_DISTMSG_REQUEST_PARAMS] = request[fwRPC_REQUEST_MODEL_PARAMS];

          fwRPC_dbg_log(fwRPC_DBG_EVENT_GENERIC,
            "Recalling " + hdl
          );

          int err = fwRPC_msgClient_sendRequest(
            ctx,
            fwRPC_DISTMSG_TYPE_REQUEST,
            msg
          );

          if (err < 0) {
            fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
              "Could not recall request: error sending message: " + err
            );
          }
        }
      } //calls

    } //ctxs
  }
}


//------------------------------------------------------------------------------
// END REDUNDANT SYSTEM MONITOR
//------------------------------------------------------------------------------
