/**
@file 
@date        30/09/2016
@author      Riku-Pekka Silvola
@copyright   CERN (BE-ICS-SDS) 

@brief Library contains SQL functions used by the unLHCLaser component.

Usage: Public

PVSS manager usage: CTRL, UI

Constraints:
- PVSS version: 3.14
- operating system: Linux.
*/

#uses "unLaser/unLaser_utils.ctl"

/// Database connection
global dbConnection g_unLHCLaser_dbConnection;

/**
* Connect to db, if the connectionName exists, then use the existing. If connexction fails return FALSE.
*
* @param sDbName   input, database name
* @param sDbUser   input, database user
* @param sDbPass   input, database password (encrypted)
*
* @return          output, true: success, false: error
*/
public bool unLaser_openDbConnection(string sDbName, string sDbUser, string sDbPass, bool printDebug = true)
{
    // Specify connection name 
    string connectionName = "unLHCLoggingDB_"+sDbName+"_"+sDbUser;

    string errTxt;
    int rc = fwDbGetConnection(connectionName, g_unLHCLaser_dbConnection);
    if (rc == -1) {
        fwDbCheckError(errTxt, g_unLHCLaser_dbConnection);
        DebugTN("ERROR Getting the connection", errTxt);
        return false;
    } else if (rc == 1) { // connection with such name does not exist yet. Create it
        string connectString = "Driver=QOCI;Database=" + sDbName + ";User=" + sDbUser + ";ENC_PWD=" + sDbPass;
        rc = fwDbOpenConnection(connectString, g_unLHCLaser_dbConnection, connectionName);
        if (rc != 0) {
          fwDbCheckError(errTxt);
          string rcTxt = (rc
                  == -1 ? "Parameter error!" :
                  rc == 1 ? "Database error! Incorrect credentials?" :
                  rc == -2 ? "Could not establish database connection!" :
                  "Undefined error");
          DebugTN("ERROR WHILE OPENING DATABASE", rc, rcTxt, errTxt);
          return false;
        }
        else if (fwDbCheckError(errTxt, g_unLHCLaser_dbConnection)) {
          DebugTN("ERROR WHILE OPENING DATABASE", errTxt);
          return false;
        }
    }
    // Set inactivity timeout to 30 seconds, after which the conection will be terminated
    rc = fwDbSetInactivityTimeout(g_unLHCLaser_dbConnection, 30);
    if (rc != 0) {
      DebugTN(rc, "Failed to set Inactivity timeout for Laser DB connection.");
    }
    //Add a debug info for existed connection
    if (printDebug) {
      mixed retVal; // note! the argument to fwDbOption must be of type mixed!
      rc = fwDbOption("Inspect", g_unLHCLaser_dbConnection, retVal);
      DebugTN(rc, retVal);
    }
    return true;
}

/**
* Get database settings of active database
*
* @param sDbName        output, database name
* @param sDbUser        output, database user
* @param sDbPass        output, database password (encrypted)
*
* @return               output, true: success, false: error
*/
public bool unLaser_getActiveDbConfig(string &sDbName, string &sDbUser, string &sDbPass) {
    
	  string sSystemName = getSystemName();
    dyn_string dbConnectionDPEs = makeDynString(sSystemName + "Laser_Info.dbConnection.Database",
                                                sSystemName + "Laser_Info.dbConnection.User",
                                                sSystemName + "Laser_Info.dbConnection.Password");
    for (unsigned i = 1; i <= dynlen(dbConnectionDPEs); i+=1) {
      	if (!dpExists(dbConnectionDPEs[i])) {
        DebugTN("[" + dbConnectionDPEs[i] + "] DPE does not exist!");
        return false;
      	}
    }
    int iRes=dpGet(
        dbConnectionDPEs[1], sDbName,
    			dbConnectionDPEs[2], sDbUser,
    			dbConnectionDPEs[3], sDbPass);

    if (iRes<0) {
      DebugTN("dpGet failed!");
      return false;
    }
    
    return true;
}

/**
* Checks for the existance of the triplets in db and if they are associated with the given source
* Possible results:
*   'Y' triplet + source exists, and source is enabled
*   'N' triplet + source exists, but source is disabled
*   ''  triplet + source does not exist
*   Anything else means triplet + source exists, but source enabled state is not valid i.e. not Y/N
*
* @param triplets   input/output, mapping of triplets to check (key: FF:FM:FC, value: source enabled state) ("Y" = yes, "N" = no, "" = triplet + source combination not found)
* @param source     input, string source id to accept
*
* @return           output, true: db query successful, false: db query resulted in an error
*/
public bool unLaser_dynCheckTripletsInDatabase(mapping &triplets, const string &source) {
  dyn_string dsExInfo;
  unsigned ALARM_ID_IDX = 1;
  unsigned SOURCE_ENABLED_IDX = 2;
  dyn_string sources;
  string errTxt, query, sDbName, sDbUser, sDbPass;
  dbCommand cmd;
  dyn_dyn_mixed data; 
  mixed retVal; 
  const dyn_string alarm_id_list = mappingKeys(triplets);
  const unsigned nof_triplets = dynlen(alarm_id_list);
  
  if (nof_triplets <= 0) {
    return true;
  }
  else {
    mapping params;
    
    for (unsigned triplet_index = 1; triplet_index <= nof_triplets; triplet_index += 1) {
      if (alarm_id_list[triplet_index] == "") {
        DebugTN("ERROR: Alarm triplet cannot be empty!");
        return false;
      }
      else {
        params[":triplet"+triplet_index] = alarm_id_list[triplet_index]; 
      }
    }        
    // Get database settings from datapoint 
    if(unLaser_getActiveDbConfig(sDbName, sDbUser, sDbPass) == false) {
      fwException_raise(dsExInfo,"ERROR", "Loading database settings failed!" ,"");      
      fwExceptionHandling_display(dsExInfo);
      return false;
    }
    else if (false == unLaser_openDbConnection(sDbName, sDbUser, sDbPass, false)) {
      fwException_raise(dsExInfo,"ERROR", "Login to database: " + sDbUser + "@" + sDbName + " failed!" ,"");
      fwExceptionHandling_display(dsExInfo);
      return false;
    }
    else {
      
      // https://its.cern.ch/jira/browse/IS-2007
      // TI section has created a unified view that contains alarm definitions for both Phoenix and Laser-lw
      // The view is accessible from both timdb (TN) and timdb-test (GPN). Laserdb is no longer supported.
      
      if (strtolower(sDbName) != "timdb" && strtolower(sDbName) != "timdb-test") {        
        DebugTN("ERROR: Non supported database '" + sDbName + "' configured. Must be timdb or timdb-test.");
        return false;        
      }
      
      query =  "SELECT ALARM_ID alarm_id, ENABLED enabled from TIMREF.VALM_LASPHO_WINCCOA";
      query += " where alarm_id IN (" + unLaser_dynConcatenate(mappingKeys(params), " , ") + ")"; // (:triplet1 , :triplet2 , :tripletN)
      query += " and SOURCE_ID = :source_id";
      
      // add source to parameters after building query to simplify code above
      params[":source_id"] = source;

      int rc = fwDbOption("SetCommandTimeout", -1, retVal);

      fwDbStartCommand(g_unLHCLaser_dbConnection, query, cmd);
      if (fwDbCheckError(errTxt, cmd)) {throw(makeError("", PRIO_SEVERE, ERR_PARAM, 54, "DB ERROR", errTxt));};

      if (mappinglen(params)) {
          fwDbBindParams(cmd, params);
          if (fwDbCheckError(errTxt, cmd)) {throw(makeError("", PRIO_SEVERE, ERR_PARAM, 54, "DB ERROR", errTxt));};
      }
      fwDbExecuteCommand(cmd);
      if (fwDbCheckError(errTxt, cmd)) {throw(makeError("", PRIO_SEVERE, ERR_PARAM, 54, "DB ERROR", errTxt));};

      fwDbGetData(cmd, data, false);
      if (fwDbCheckError(errTxt, cmd)) {throw(makeError("", PRIO_SEVERE, ERR_PARAM, 54, "DB ERROR", errTxt));};

      fwDbFinishCommand(cmd);
      if (fwDbCheckError(errTxt)) {throw(makeError("", PRIO_SEVERE, ERR_PARAM, 54, "DB ERROR", errTxt));};
    }
    
    DebugFTN("echoLaserdbTriplets", data);
    
    const unsigned data_len = dynlen(data);
    for (unsigned data_idx = 1; data_idx <= data_len; data_idx += 1) {
      if (dynlen(data[data_idx]) < SOURCE_ENABLED_IDX) {
        DebugTN("Database responses number of columns mismatch from expected! [" + dynlen(data[data_idx]) + "/" + SOURCE_ENABLED_IDX + "] Check query.");
        return false;
      }
      else if (!mappingHasKey(triplets, data[data_idx][ALARM_ID_IDX])) {
        DebugTN("Database responses included unexpected alarm_ids! [" + data[data_idx][ALARM_ID_IDX] + "] Check query.");
        return false;
      }
      else {
        triplets[data[data_idx][ALARM_ID_IDX]] = (data[data_idx][SOURCE_ENABLED_IDX] != "" ? data[data_idx][SOURCE_ENABLED_IDX] : "null");
      }
    }
  }
  return true;
}
